﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Encodings.PackingEncoding.MultiComponentVector {
  [Item("Exhaustive MultiComponentVector Move Generator", "Generates all possible multi component moves from a given multiComponentVector.")]
  [StorableClass]
  public class ExhaustiveMultiComponentVectorMoveGenerator : MultiComponentVectorMoveGenerator, IExhaustiveMoveGenerator {
    [StorableConstructor]
    protected ExhaustiveMultiComponentVectorMoveGenerator(bool deserializing) : base(deserializing) { }
    protected ExhaustiveMultiComponentVectorMoveGenerator(ExhaustiveMultiComponentVectorMoveGenerator original, Cloner cloner) : base(original, cloner) { }
    public ExhaustiveMultiComponentVectorMoveGenerator() : base() { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new ExhaustiveMultiComponentVectorMoveGenerator(this, cloner);
    }


    public static MultiComponentVectorMove[] Apply(MultiComponentVectorEncoding multiComponentVector) {
      //NrOfMoves = nrOfRotationMoves + nrOfSingleGroupingMoves + nrOfSwapSequenceMoves
      //  nrOfRotationMoves       = nrOfItems
      //  nrOfSingleGroupingMoves = nrOfItems * (nrOfBins - 1)
      //  nrOfSwapSequenceMoves   = SUM(foreach binNr: !nrOfItemsInBin)

      int lastBinNr = 0;
      foreach (var pi in multiComponentVector.PackingInformations) {
        if (pi.AssignedBin > lastBinNr)
          lastBinNr = pi.AssignedBin;
      }
      int nrOfBins = lastBinNr + 1;

      int nrOfRotationMoves = multiComponentVector.PackingInformations.Count;
      int nrOfSingleGroupingMoves = nrOfRotationMoves * (nrOfBins - 1);

      int[] itemsPerBin = new int[nrOfBins];
      foreach (var pi in multiComponentVector.PackingInformations) {
        itemsPerBin[pi.AssignedBin]++;
      }
      int nrOfSwapSequenceMoves = 0;
      foreach (int items in itemsPerBin) {
        nrOfSwapSequenceMoves += (items * (items - 1)) / 2;
      }

      int totalMoves = nrOfRotationMoves + nrOfSingleGroupingMoves + nrOfSwapSequenceMoves;
      MultiComponentVectorMove[] moves = new MultiComponentVectorMove[totalMoves];
      int count = 0;
      foreach (MultiComponentVectorMove move in GenerateRotationMoves(multiComponentVector)) {
        moves[count++] = move;
      }
      foreach (MultiComponentVectorMove move in GenerateSingleGroupingMoves(multiComponentVector)) {
        moves[count++] = move;
      }
      foreach (MultiComponentVectorMove move in GenerateSwapSequenceMoves(multiComponentVector)) {
        moves[count++] = move;
      }
      return moves;
    }    

    protected override MultiComponentVectorMove[] GenerateMoves(MultiComponentVectorEncoding multiComponentVector) {
      return Apply(multiComponentVector);
    }
  }
}
