﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.BinPacking.Interfaces;

namespace HeuristicLab.Encodings.PackingEncoding.MultiComponentVector {
  [Item("MultiComponentVectorMoveGenerator", "Base class for all multi-component vector move generators.")]
  [StorableClass]
  public abstract class MultiComponentVectorMoveGenerator : SingleSuccessorOperator, IMultiComponentVectorMoveOperator, IMoveGenerator {
    public override bool CanChangeName {
      get { return false; }
    }
    public ILookupParameter<MultiComponentVectorEncoding> MultiComponentVectorParameter {
      get { return (ILookupParameter<MultiComponentVectorEncoding>)Parameters["MultiComponentVector"]; }
    }
    public ILookupParameter<MultiComponentVectorMove> MultiComponentVectorMoveParameter {
      get { return (LookupParameter<MultiComponentVectorMove>)Parameters["MultiComponentVectorMove"]; }
    }
    protected ScopeParameter CurrentScopeParameter {
      get { return (ScopeParameter)Parameters["CurrentScope"]; }
    }

    [StorableConstructor]
    protected MultiComponentVectorMoveGenerator(bool deserializing) : base(deserializing) { }
    protected MultiComponentVectorMoveGenerator(MultiComponentVectorMoveGenerator original, Cloner cloner) : base(original, cloner) { }
    public MultiComponentVectorMoveGenerator()
      : base() {
        Parameters.Add(new LookupParameter<MultiComponentVectorEncoding>("MultiComponentVector", "The multi component vector for which moves should be generated."));
      Parameters.Add(new LookupParameter<MultiComponentVectorMove>("MultiComponentVectorMove", "The moves that should be generated in subscopes."));
      Parameters.Add(new ScopeParameter("CurrentScope", "The current scope where the moves should be added as subscopes."));
    }




    public static IEnumerable<MultiComponentVectorMove> GenerateRotationMoves(MultiComponentVectorEncoding multiComponentVector) {
      for (int i = 0; i < multiComponentVector.PackingInformations.Count; i++)
        yield return new SingleItemRotationMove(i, multiComponentVector);
    }

    public static IEnumerable<MultiComponentVectorMove> GenerateSingleGroupingMoves(MultiComponentVectorEncoding multiComponentVector) {
      int nrOfBins = 0;
      foreach (var pi in multiComponentVector.PackingInformations) {
        if (pi.AssignedBin > nrOfBins)
          nrOfBins = pi.AssignedBin;
      }
      nrOfBins++;
      for (int i = 0; i < multiComponentVector.PackingInformations.Count; i++)
        for (int j = 0; j < nrOfBins; j++) {
          if (j != multiComponentVector.PackingInformations[i].AssignedBin)
            yield return new SingleGroupingMove(i, j, multiComponentVector);
        }
    }

    public static IEnumerable<MultiComponentVectorMove> GenerateSwapSequenceMoves(MultiComponentVectorEncoding multiComponentVector) {
      Dictionary<int, List<int>> indexesPerBin = new Dictionary<int, List<int>>();
      for (int i = 0; i < multiComponentVector.PackingInformations.Count; i++) {
        int currentBin = multiComponentVector.PackingInformations[i].AssignedBin;
        if (!indexesPerBin.ContainsKey(currentBin))
          indexesPerBin[currentBin] = new List<int>();
        indexesPerBin[currentBin].Add(i);
      }
      foreach (var entry in indexesPerBin) {
        for (int i = 0; i < entry.Value.Count - 1; i++)
          for (int j = i + 1; j < entry.Value.Count; j++)
            yield return new SwapSequenceMove(entry.Value[i], entry.Value[j], multiComponentVector);
      }
    }


    public override IOperation Apply() {
      MultiComponentVectorEncoding mcv = MultiComponentVectorParameter.ActualValue;
      MultiComponentVectorMove[] moves = GenerateMoves(mcv);
      Scope[] moveScopes = new Scope[moves.Length];
      for (int i = 0; i < moveScopes.Length; i++) {
        moveScopes[i] = new Scope(i.ToString());
        moveScopes[i].Variables.Add(new Variable(MultiComponentVectorMoveParameter.ActualName, moves[i]));
      }
      CurrentScopeParameter.ActualValue.SubScopes.AddRange(moveScopes);
      return base.Apply();
    }

    protected abstract MultiComponentVectorMove[] GenerateMoves(MultiComponentVectorEncoding multiComponentVector);
  }
}
